import pytest

from ytpb.utils.path import (
    adjust_title_for_filename,
    format_title_for_filename,
    TitleAllowedCharacters,
)


@pytest.mark.parametrize(
    "title,expected",
    [
        (
            "FRANCE 24 — EN DIRECT — Info et actualités internationales en continu 24h/24",  # noqa: E501
            "FRANCE 24 — EN DIRECT — Info et actualités internationales en continu 24h-24",  # noqa: E501
        ),
        (
            "🛑 LE JOURNAL TÉLÉVISÉ DE 20H - VENDREDI 24 NOVEMBRE 2023",
            "🛑 LE JOURNAL TÉLÉVISÉ DE 20H - VENDREDI 24 NOVEMBRE 2023",
        ),
        (
            "【LIVE】新宿駅前の様子 Shinjuku, Tokyo JAPAN【ライブカメラ】 | TBS NEWS DIG",  # noqa: E501
            "【LIVE】新宿駅前の様子 Shinjuku, Tokyo JAPAN【ライブカメラ】 | TBS NEWS DIG",  # noqa: E501
        ),
    ],
)
def test_adjust_title_for_filename_as_is(title, expected):
    assert expected == adjust_title_for_filename(title, separator=None, max_length=None)


@pytest.mark.parametrize(
    "title,expected",
    [
        (
            "FRANCE 24 — EN DIRECT — Info et actualités internationales en continu 24h/24",  # noqa: E501
            "FRANCE-24—EN-DIRECT—Info-et-actualités-internationales-en-continu-24h-24",
        ),
        (
            "🛑 LE JOURNAL TÉLÉVISÉ DE 20H - VENDREDI 24 NOVEMBRE 2023",
            "🛑-LE-JOURNAL-TÉLÉVISÉ-DE-20H-VENDREDI-24-NOVEMBRE-2023",
        ),
        (
            "【LIVE】新宿駅前の様子 Shinjuku, Tokyo JAPAN【ライブカメラ】 | TBS NEWS DIG",  # noqa: E501
            "【LIVE】新宿駅前の様子-Shinjuku,-Tokyo-JAPAN【ライブカメラ】-|-TBS-NEWS-DIG",  # noqa: E501
        ),
    ],
)
def test_adjust_title_for_filename_with_separator(title, expected):
    assert expected == adjust_title_for_filename(title, separator="-", max_length=None)


@pytest.mark.parametrize(
    "title,expected",
    [
        (
            "FRANCE 24 — EN DIRECT — Info et actualités internationales en continu 24h/24",  # noqa: E501
            "FRANCE 24 -- EN DIRECT -- Info et actualites internationales en continu 24h-24",  # noqa: E501
        ),
        (
            "🛑 LE JOURNAL TÉLÉVISÉ DE 20H - VENDREDI 24 NOVEMBRE 2023",
            "LE JOURNAL TELEVISE DE 20H - VENDREDI 24 NOVEMBRE 2023",
        ),
        (
            "【LIVE】新宿駅前の様子 Shinjuku, Tokyo JAPAN【ライブカメラ】 | TBS NEWS DIG",  # noqa: E501
            "[(LIVE)] Xin Su Yi Qian noYang Zi Shinjuku, Tokyo JAPAN[(raibukamera)] | TBS NEWS DIG",  # noqa: E501
        ),
    ],
)
def test_adjust_title_for_filename_with_ascii_and_no_separator(title, expected):
    assert expected == adjust_title_for_filename(
        title, separator=None, max_length=None, characters=TitleAllowedCharacters.ASCII
    )


@pytest.mark.parametrize(
    "title,expected",
    [
        (
            "FRANCE 24 — EN DIRECT — Info et actualités internationales en continu 24h/24",  # noqa: E501
            "FRANCE-24--EN-DIRECT--Info-et-actualites-internationales-en-continu-24h-24",  # noqa: E501
        ),
        (
            "🛑 LE JOURNAL TÉLÉVISÉ DE 20H - VENDREDI 24 NOVEMBRE 2023",
            "LE-JOURNAL-TELEVISE-DE-20H-VENDREDI-24-NOVEMBRE-2023",
        ),
        (
            "【LIVE】新宿駅前の様子 Shinjuku, Tokyo JAPAN【ライブカメラ】 | TBS NEWS DIG",  # noqa: E501
            "[(LIVE)]-Xin-Su-Yi-Qian-noYang-Zi-Shinjuku,-Tokyo-JAPAN[(raibukamera)]-|-TBS-NEWS-DIG",  # noqa: E501
        ),
    ],
)
def test_adjust_title_for_filename_with_ascii_and_separator(title, expected):
    assert expected == adjust_title_for_filename(
        title, separator="-", max_length=None, characters=TitleAllowedCharacters.ASCII
    )


@pytest.mark.parametrize(
    "title,expected,separator",
    [
        (
            "Jackman Maine - Newton Field Airport - Cam Left",
            "Jackman-Maine-Newton-Field-Airport-Cam-Left",
            "-",
        ),
        (
            "Jackman Maine -  Newton Field Airport  -  Cam Left",
            "Jackman-Maine-Newton-Field-Airport-Cam-Left",
            "-",
        ),
        (
            "Jackman Maine —  Newton Field Airport - Cam Left",
            "Jackman-Maine--Newton-Field-Airport-Cam-Left",
            "-",
        ),
        (
            "Jackman Maine - Newton Field Airport - Cam Left",
            "Jackman_Maine-Newton_Field_Airport-Cam_Left",
            "_",
        ),
        (
            "Jackman Maine -  Newton Field Airport  -  Cam Left",
            "Jackman_Maine-Newton_Field_Airport-Cam_Left",
            "_",
        ),
        (
            "Jackman Maine —  Newton Field Airport - Cam Left",
            "Jackman_Maine--Newton_Field_Airport-Cam_Left",
            "_",
        ),
    ],
)
def test_corner_cases_of_adjust_title_for_filename(title, expected, separator):
    assert expected == adjust_title_for_filename(
        title, separator=separator, characters=TitleAllowedCharacters.ASCII
    )
    assert expected == adjust_title_for_filename(
        title, separator=separator, characters=TitleAllowedCharacters.POSIX
    )


def test_adjust_title_for_filename_with_posix():
    title = "【LIVE】新宿駅前の様子 Shinjuku, Tokyo JAPAN | TBS NEWS DIG"

    expected = "LIVE-Xin-Su-Yi-Qian-noYang-Zi-Shinjuku-Tokyo-JAPAN-TBS-NEWS-DIG"
    assert expected == adjust_title_for_filename(
        title, separator=None, max_length=None, characters=TitleAllowedCharacters.POSIX
    )

    expected = "LIVE_Xin_Su_Yi_Qian_noYang_Zi_Shinjuku_Tokyo_JAPAN_TBS_NEWS_DIG"
    assert expected == adjust_title_for_filename(
        title, separator="_", max_length=None, characters=TitleAllowedCharacters.POSIX
    )

    expected = "LIVE-Xin-Su-Yi-Qian-noYang-Zi-Shinjuku-Tokyo-JAPAN-TBS-NEWS-DIG"
    assert expected == adjust_title_for_filename(
        title, separator="*", max_length=None, characters=TitleAllowedCharacters.POSIX
    )


@pytest.mark.parametrize(
    "title,expected,max_length",
    [
        (  # 1                                        42
            "FRANCE 24 — EN DIRECT — Info et actualités",
            "FRANCE 24 — EN DIRECT — Info et actualités",
            42,
        ),
        (  # 1                   21
            "FRANCE 24 — EN DIRECT — Info et actualités",
            "FRANCE 24 — EN DIRECT",
            21,
        ),
        (  # 1                  20
            "FRANCE 24 — EN DIRECT — Info et actualités",
            "FRANCE 24 — EN",
            20,
        ),
        (  # 1           13
            "FRANCE 24 — EN DIRECT — Info et actualités",
            "FRANCE 24",
            13,
        ),
    ],
)
def test_adjust_title_for_filename_with_max_length(title, expected, max_length):
    assert expected == adjust_title_for_filename(
        title, separator=None, max_length=max_length
    )


@pytest.mark.parametrize(
    "title,expected,max_length",
    [
        (
            "FRANCE 24—EN DIRECT—Info et actualités",
            #  1                        26
            # "FRANCE-24--EN-DIRECT--Info-et-actualités",
            "FRANCE-24--EN-DIRECT--Info",
            26,
        ),
        (
            "FRANCE 24—EN DIRECT—Info et actualités",
            #  1                       24
            # "FRANCE-24--EN-DIRECT--Info-et-actualités",
            "FRANCE-24--EN-DIRECT",
            24,
        ),
    ],
)
def test_adjust_title_for_filename_with_max_length_and_separator(
    title, expected, max_length
):
    assert expected == adjust_title_for_filename(
        title,
        separator="-",
        max_length=max_length,
        characters=TitleAllowedCharacters.ASCII,
    )


def test_format_title_for_filename_with_original_style():
    title = "【LIVE】新宿駅前の様子 Shinjuku, Tokyo JAPAN | TBS NEWS DIG"
    assert title == format_title_for_filename(title, style="original")
    assert title == format_title_for_filename(
        title, style="original", separator="-", lowercase=True
    )


@pytest.mark.parametrize(
    "title,expected,characters",
    [
        (
            "【LIVE】新宿駅前の様子 Shinjuku, Tokyo JAPAN | TBS NEWS DIG",
            "【LIVE】新宿駅前の様子 Shinjuku, Tokyo JAPAN | TBS NEWS DIG",
            TitleAllowedCharacters.UNICODE,
        ),
        (
            "【LIVE】新宿駅前の様子 Shinjuku, Tokyo JAPAN | TBS NEWS DIG",
            "[(LIVE)] Xin Su Yi Qian noYang Zi Shinjuku, Tokyo JAPAN | TBS NEWS DIG",
            TitleAllowedCharacters.ASCII,
        ),
        (
            "【LIVE】新宿駅前の様子 Shinjuku, Tokyo JAPAN | TBS NEWS DIG",
            "LIVE-Xin-Su-Yi-Qian-noYang-Zi-Shinjuku-Tokyo-JAPAN-TBS-NEWS-DIG",
            TitleAllowedCharacters.POSIX,
        ),
    ],
)
def test_format_title_for_filename_with_characters(title, expected, characters):
    assert expected == format_title_for_filename(
        title, style="custom", characters=characters
    )



import enum
import re
import string
import textwrap
import unicodedata
import warnings
from datetime import datetime, timedelta
from pathlib import Path
from typing import Callable, TypedDict

import pathvalidate
import unidecode
from timedelta_isoformat import timedelta as isotimedelta

from ytpb.config import AddressableDict
from ytpb.types import AddressableMappingProtocol
from ytpb.utils.date import ISO8601DateFormatter


DASHES = [
    "\u002D",  # HYPHEN-MINUS
    "\u2010",  # HYPHEN
    "\u2011",  # NON-BREAKING HYPHEN
    "\u2012",  # FIGURE DASH
    "\u2013",  # EN DASH
    "\u2014",  # EM DASH
    "\u2015",  # HORIZONTAL BAR
]

OUTPUT_PATH_PLACEHOLDER_NAME = r"[a-z_]+"
OUTPUT_PATH_PLACEHOLDER_RE = re.compile(rf"<({OUTPUT_PATH_PLACEHOLDER_NAME})>")


class TitleAllowedCharacters(enum.StrEnum):
    UNICODE = enum.auto()
    ASCII = enum.auto()
    POSIX = enum.auto()


class ChevronTemplate(string.Template):
    """A template that supports <variable>-style substitution."""

    delimiter = "<"
    pattern = rf"""
        <(?:
            (?P<escaped><) |
            (?P<named>{OUTPUT_PATH_PLACEHOLDER_NAME})> |
            (?P<braced>) |
            (?P<invalid>)
        )
    """


class OutputPathContextRenderer:
    @staticmethod
    def render_title(value: str, settings: AddressableMappingProtocol) -> str:
        style_address = "output.title.style"
        style = settings.traverse(style_address)
        match style:
            case "original" | None:
                output = format_title_for_filename(value)
            case "custom":
                custom_settings = AddressableDict(
                    settings.traverse("output.title.custom")
                )
                output = format_title_for_filename(
                    value, style=style, **custom_settings
                )
            case _:
                warnings.warn(
                    f"Unknown style '{style}', fallback to 'original'. "
                    f"Check the {style_address} settings"
                )
                output = format_title_for_filename(value, style="original")
        return output

    @staticmethod
    def render_date(
        value: datetime, formatter, settings: AddressableMappingProtocol
    ) -> str:
        output_date_styles = settings.traverse("output.date.styles")
        if output_date_styles:
            return formatter.format(f"{{:{output_date_styles}}}", value)
        else:
            return formatter.format("{}", value)

    @staticmethod
    def render_duration(value: timedelta, settings: AddressableMappingProtocol) -> str:
        seconds_rounded = round(value.total_seconds())
        return isotimedelta(seconds=seconds_rounded).isoformat()


class MinimalOutputPathContext(TypedDict):
    id: str
    title: str


class IntervalOutputPathContext(TypedDict):
    input_start_date: datetime
    input_end_date: datetime
    actual_start_date: datetime
    actual_end_date: datetime
    duration: timedelta


def render_minimal_output_path_context(
    context: MinimalOutputPathContext,
    config_settings: AddressableMappingProtocol | None = None,
) -> MinimalOutputPathContext:
    config_settings = config_settings or AddressableMappingProtocol({})
    output: MinimalOutputPathContext = {}
    for variable in MinimalOutputPathContext.__annotations__.keys():
        match variable:
            case "title" as x:
                output[x] = OutputPathContextRenderer.render_title(
                    context[x], config_settings
                )
            case _ as x:
                output[x] = context[x]
    return output


def render_interval_output_path_context(
    context: IntervalOutputPathContext,
    config_settings: AddressableMappingProtocol | None = None,
) -> IntervalOutputPathContext:
    config_settings = config_settings or AddressableMappingProtocol({})

    date_formatter = ISO8601DateFormatter()

    output: IntervalOutputPathContext = {}
    for variable in IntervalOutputPathContext.__annotations__.keys():
        match variable:
            case x if x.endswith("_date"):
                output[x] = OutputPathContextRenderer.render_date(
                    context[x], date_formatter, config_settings
                )
            case "duration" as x:
                output[x] = OutputPathContextRenderer.render_duration(
                    context[x], config_settings
                )
            case _ as x:
                output[x] = context[x]

    return output


def expand_template_output_path(
    path: Path,
    template_context: dict,
    render_function: Callable[[dict, AddressableMappingProtocol], dict],
    config_settings: AddressableMappingProtocol | None = None,
) -> Path:
    if config_settings is None:
        config_settings = AddressableDict({})

    path_string = str(path)

    matched_variables = OUTPUT_PATH_PLACEHOLDER_RE.findall(path_string)
    known_variables = set(template_context.keys())
    if unknown_variables := set(matched_variables) - known_variables:
        warnings.warn(f"Unknown variables: {unknown_variables}")

    rendered_context = render_function(template_context, config_settings)
    output = ChevronTemplate(path_string).substitute(rendered_context)

    return Path(output)


def sanitize_filename(value: str) -> str:
    normalized = unicodedata.normalize("NFKC", value)
    output = pathvalidate.sanitize_filename(
        normalized, replacement_text="-", platform="POSIX"
    )
    return output


def posixify_filename(value: str, separator: str = "-"):
    posix_characters_re = re.compile(r"[^\w\-.]+", flags=re.ASCII)

    if not posix_characters_re.match(separator):
        actual_separator = separator
    else:
        actual_separator = "-"

    output = unidecode.unidecode(value, "replace", " ")
    output = posix_characters_re.sub(actual_separator, output)
    # Strip the space if a title started with a non-POSIX character.
    output = output.lstrip(actual_separator)

    return output


def adjust_title_for_filename(
    title: str,
    separator: str | None = None,
    max_length: int | None = None,
    characters: TitleAllowedCharacters = TitleAllowedCharacters.UNICODE,
) -> str:
    fallback_separator = "-"
    dashes_pattern = r"(?:\s+)?([{0}]+)(?:\s+)?".format("".join(DASHES))

    if separator:
        # For visual appeal, replace dashes and surrounding spaces with
        # dashes. Compare, e.g.: 'A_B_-_C_D' and 'A_B-C_D' (we prefer this one).
        title = re.sub(dashes_pattern, r"\1", title)

    if characters == TitleAllowedCharacters.POSIX:
        output = posixify_filename(title, separator or fallback_separator)
    else:
        output = sanitize_filename(title)

        actual_separator: str

        match characters:
            case TitleAllowedCharacters.UNICODE:
                actual_separator = separator
            case TitleAllowedCharacters.ASCII:
                actual_separator = separator and unidecode.unidecode(
                    separator, "replace", fallback_separator
                )
                output = unidecode.unidecode(output, "replace", " ")
                # Strip the space if a title started with a non-converted character.
                output = output.lstrip()

        # Replace multiple consecutive spaces with a single space. This can be
        # in an original title, as well as after converting a title using ASCII
        # codec (non-converted characters are replaced with a space).
        output = re.sub(r"\s+", " ", output)

    if max_length and len(output) > max_length:
        output = textwrap.shorten(output, max_length, placeholder="")
        output = re.sub(dashes_pattern + "$", "", output)

    if separator and characters != TitleAllowedCharacters.POSIX:
        # For the POSIX case, the output is already with spaces replaced after
        # the posixify_filename() function.
        output = re.sub(r"\s", actual_separator, output)

    return output


def format_title_for_filename(
    title: str,
    style: str = "original",
    separator: str | None = None,
    max_length: int | None = None,
    characters: TitleAllowedCharacters = TitleAllowedCharacters.UNICODE,
    lowercase: bool = False,
):
    match style:
        case "original":
            output = sanitize_filename(title)
        case "custom":
            output = adjust_title_for_filename(
                title, separator, max_length, characters=characters
            )
            if lowercase:
                output = output.lower()
        case _:
            raise ValueError(f"Unknown title formatting style: '{style}'")

    return output



import re
from typing import Any, Callable

import pycond as pc

import structlog

from ytpb.config import ALIAS_EXPAND_FUNCTIONS, ALL_ALIASES
from ytpb.exceptions import QueryError
from ytpb.types import AudioOrVideoStream

logger = structlog.get_logger(__name__)

ALIAS_RE = re.compile(r"@([\w<>=-]+)(?!\s)?")
FORMAT_SPEC_RE = re.compile(
    r"^(?:(?P<function>[\w\-]+)\((?P<expr>[^\(\)]+)\)|(?P<just_expr>[^\(\)]+))$"
)


def _expand_aliases(expression: str, aliases: dict[str, str]) -> str:
    for f in ALIAS_EXPAND_FUNCTIONS:
        expression = f(expression)

    all_aliases = ALL_ALIASES | aliases
    for matched in ALIAS_RE.finditer(expression):
        alias_with_symbol = matched.group()
        alias = matched.group(1)
        try:
            aliased_value = all_aliases[alias]
            expression = expression.replace(alias_with_symbol, aliased_value)
        except KeyError:
            raise QueryError(
                f"Unknown alias in format spec: '{alias_with_symbol}'"
            ) from None
    return expression


def treat_none_as_false(operator_function: Callable, a: Any, b: Any) -> bool:
    if a is None:
        return False
    else:
        return operator_function(a, b)


def custom_lookup(
    attribute: str, target: str | int, cfg: dict, state: AudioOrVideoStream, **kwargs
) -> tuple[str, str | int]:
    """Custom lookup function for audio and video streams (representations) with
    the existence check of attributes."""
    try:
        if isinstance(attribute, tuple):
            raise QueryError(
                "Attribute incorrectly matched in format spec: '{}'".format(
                    " ".join(attribute)
                )
            )
        attribute_value = getattr(state, attribute)
    except AttributeError as e:
        raise QueryError(f"Unknown attribute in format spec: '{attribute}'") from e
    return attribute_value, target


def make_filter_from_expression(expression: str, aliases: dict[str, str] | None = None):
    if ALIAS_RE.search(expression):
        expression = _expand_aliases(expression, aliases or {})
        logger.debug(f"Expression with alias(es) expanded as '{expression}'")
    output = pc.make_filter(
        expression,
        autoconv_lookups=True,
        lookup=custom_lookup,
        ops_thru=treat_none_as_false,
    )
    return output



import email
import sys
import textwrap
from datetime import datetime, timedelta
from pathlib import Path

import click
import cloup
import structlog

from ytpb.download import compose_default_segment_filename
from ytpb.exceptions import (
    BadCommandArgument,
    BaseUrlExpiredError,
    BroadcastStatusError,
    CachedItemNotFoundError,
    QueryError,
)
from ytpb.fetchers import YoutubeDLInfoFetcher, YtpbInfoFetcher
from ytpb.info import BroadcastStatus
from ytpb.playback import Playback, RewindInterval
from ytpb.segment import Segment
from ytpb.types import DateInterval, SetOfStreams
from ytpb.utils.date import format_timedelta, round_date
from ytpb.utils.url import extract_parameter_from_url, normalize_video_url

logger = structlog.getLogger(__name__)


CONSOLE_TEXT_WIDTH = 80
EARLIEST_DATE_TIMEDELTA = timedelta(days=6, hours=23)


def get_parameter_by_name(name: str, ctx: click.Context) -> click.Parameter:
    return next((p for p in ctx.command.params if p.name == name), None)


def create_playback(ctx: click.Context) -> Playback:
    stream_url = ctx.params["stream_url"]

    if ctx.params["yt_dlp"]:
        fetcher = YoutubeDLInfoFetcher(stream_url)
    else:
        fetcher = YtpbInfoFetcher(stream_url)

    try:
        if from_manifest := ctx.params.get("from_manifest"):
            try:
                click.echo("Run playback from manifest file")
                playback = Playback.from_manifest(from_manifest, fetcher=fetcher)
            except BaseUrlExpiredError:
                click.echo("Oh no, the manifest has been expired, exit.", err=True)
                sys.exit(1)
        else:
            click.echo(f"Run playback for {stream_url}")
            click.echo("(<<) Collecting info about the video...")

            force_update_cache = ctx.params.get("force_update_cache", False)
            need_read_cache = not (
                ctx.params.get("no_cache", True) or force_update_cache
            )
            if need_read_cache:
                try:
                    playback = Playback.from_cache(stream_url, fetcher=fetcher)
                except CachedItemNotFoundError:
                    logger.debug("Could not find unexpired cached item for the video")
                    playback = Playback.from_url(
                        stream_url, fetcher=fetcher, write_to_cache=True
                    )
            else:
                playback = Playback.from_url(
                    stream_url, fetcher=fetcher, write_to_cache=force_update_cache
                )
    except BroadcastStatusError as e:
        match e.status:
            case BroadcastStatus.NONE:
                click.echo("It's seems that the video is not a live stream", err=True)
            case BroadcastStatus.COMPLETED:
                click.echo("Stream was live, but now it's finished", err=True)
        sys.exit(1)

    click.echo(f"Stream '{playback.info.title}' is alive!")

    return playback


def normalize_stream_url(ctx: click.Context, param: click.Argument, value: str) -> str:
    try:
        return normalize_video_url(value)
    except BadCommandArgument as e:
        raise click.BadParameter(str(e))


stream_argument = cloup.argument(
    "stream_url",
    metavar="STREAM",
    help="Stream URL or video ID.",
    callback=normalize_stream_url,
)


def check_end_options(start, end, duration, preview):
    if not (end or duration or preview):
        raise click.UsageError(
            "One of --end, --duration, or --preview must be specified."
        )

    if isinstance(end, datetime) and isinstance(start, datetime):
        if end <= start:
            raise click.BadParameter(
                "End date is ahead or equal to the start date.",
                param_hint="'-e' / '--end'",
            )


def get_downloaded_segment(playback, sequence, base_url):
    segment_filename = compose_default_segment_filename(sequence, base_url)
    segment = Segment.from_file(playback.get_temp_directory() / segment_filename)
    return segment


def prepare_line_for_summary_info(
    date: datetime,
    input_actual_timedelta: timedelta,
    use_ms_precision: bool = False,
) -> str:
    if not use_ms_precision:
        date = round_date(date)
    date_diff_string = format_timedelta(input_actual_timedelta, use_ms_precision)

    formatted = email.utils.format_datetime(date).split(", ")[1]
    if use_ms_precision:
        date_part, tz_part = formatted[:-6], formatted[-5:]
        ms_value = f"{date.microsecond // 1000:03g}"
        formatted = f"{date_part}.{ms_value} {tz_part}"

    if date_diff_string[1:] != "00:00":
        output = f"{formatted} ({date_diff_string})"
    else:
        output = formatted

    return output


def print_summary_info(
    input_date_interval: DateInterval,
    actual_date_interval: DateInterval,
    rewind_interval: RewindInterval,
    use_ms_precision: bool = False,
) -> None:
    input_tzinfo = input_date_interval.start.tzinfo

    start_time_info_line = prepare_line_for_summary_info(
        actual_date_interval.start.astimezone(input_tzinfo),
        actual_date_interval.start - input_date_interval.start,
        use_ms_precision,
    )
    click.echo(
        "Actual start: {}, seq. {}".format(
            start_time_info_line, rewind_interval.start.sequence
        )
    )

    end_time_info_line = prepare_line_for_summary_info(
        actual_date_interval.end.astimezone(input_tzinfo),
        actual_date_interval.end - input_date_interval.end,
        use_ms_precision,
    )
    click.echo(
        "  Actual end: {}, seq. {}".format(
            end_time_info_line, rewind_interval.end.sequence
        )
    )

    # total_duration = format_duration(
    #     timedelta(seconds=actual_date_interval.duration),
    #     DurationFormatPattern.IN_SENTENCE
    # )
    # click.echo(f"Total duration: {total_duration}, estimated size: ?")


def check_streams_not_empty(
    audio_stream, audio_format, video_stream, video_format, max_text_width=62
):
    if audio_format and not audio_stream:
        message = (
            "  Error! Found no audio formats matching requested format spec: "
            f"{audio_format}."
        )
        if video_format and not video_stream:
            message += " Same for the video: {video_format}."
        click.echo(textwrap.fill(message, max_text_width))
        click.echo("~" * max_text_width)
        sys.exit(1)
    elif video_format and not video_stream:
        click.echo("~" * max_text_width)
        message = (
            "  Error! Found no video formats matching requested format spec: "
            f"{video_format}."
        )
        click.echo(textwrap.fill(message, max_text_width))
        click.echo("~" * max_text_width)
        sys.exit(1)


def query_streams_or_exit(
    streams: SetOfStreams,
    format_spec: str,
    param: str | None = None,
    allow_empty: bool = False,
    allow_many: bool = True,
) -> SetOfStreams:
    try:
        queried_streams = streams.query(format_spec)
    except QueryError as exc:
        message = f"error: Invalid value for '{param}'. {exc}: '{format_spec}'"
        click.echo(message, err=True)
        sys.exit(1)

    if queried_streams:
        if len(queried_streams) > 1 and not allow_many:
            logger.error(
                "Too many queried streams", itags=[s.itag for s in queried_streams]
            )
            click.echo(
                (
                    "error: Format spec is ambiguous: '{format_spec}'.\n\n"
                    "Found more than one stream matched a format spec. Please be more "
                    "explicit, or try 'yt-dlp --live-from-start -F ...' first."
                )
            )
            sys.exit(1)
    elif not allow_empty:
        message = (
            f"error: No streams found matching '{param}' format spec: '{format_spec}'"
        )
        click.echo(message, err=True)
        sys.exit(1)

    return queried_streams


def raise_for_too_far_sequence(
    sequence: int,
    current_sequence: int,
    base_url: str,
    ctx: click.Context,
    param_name: str,
) -> None:
    segment_duration = int(float(extract_parameter_from_url("dur", base_url)))
    earliest_sequence = (
        current_sequence - EARLIEST_DATE_TIMEDELTA.total_seconds() // segment_duration
    )
    if sequence <= earliest_sequence:
        click.echo("\n")
        raise click.BadParameter(
            f"Sequence number is beyond the limit of 7 days: {sequence}",
            param=get_parameter_by_name(param_name, ctx),
        )


def raise_for_sequence_ahead_of_current(
    sequence, current_sequence: int, ctx: click.Context, param_name: str
) -> None:
    if sequence >= current_sequence:
        click.echo("\n")
        message = (
            "Sequence number is ahead or equal to the current one: "
            f"{sequence} >= {current_sequence}"
        )
        param = get_parameter_by_name(param_name, ctx)
        raise click.BadParameter(message, ctx, param)


def resolve_output_path(output_path: Path) -> Path:
    resolved_path = Path(output_path).expanduser().resolve()
    resolved_path.parent.mkdir(parents=True, exist_ok=True)
    return resolved_path



from collections.abc import MutableSet
from operator import attrgetter
from typing import Any, Callable, Iterator, Self

from ytpb.conditional import FORMAT_SPEC_RE, make_filter_from_expression
from ytpb.exceptions import QueryError
from ytpb.types import AudioOrVideoStream, AudioStream, SetOfStreams, VideoStream


def stream_comparison_function(stream: AudioOrVideoStream):
    match stream:
        case AudioStream():
            return (stream.mime_type, stream.audio_sampling_rate)
        case VideoStream():
            return (stream.mime_type, stream.height, stream.frame_rate)


class Streams(MutableSet):
    """Represents a set of `ytpb.info.RepresentationInfo` objects."""

    def __init__(self, iterable: list[AudioOrVideoStream] | None = None) -> None:
        self._elements = set()
        for value in iterable or []:
            if not isinstance(value, AudioOrVideoStream):
                raise ValueError
            self._elements.add(value)

    @classmethod
    def from_dicts(cls, dicts: list[dict]):
        streams = cls()
        stream: AudioOrVideoStream
        for stream_dict in dicts:
            if "audio" in stream_dict["mime_type"]:
                stream = AudioStream(**stream_dict)
            else:
                stream = VideoStream(**stream_dict)
            streams.add(stream)
        return streams

    def __len__(self):
        return len(self._elements)

    def __ior__(self, other: Self) -> Self:
        self.update(other)
        return self

    def __iter__(self) -> Iterator[AudioOrVideoStream]:
        return iter(self._elements)

    def __contains__(self, item: Any) -> bool:
        for stream in self:
            if stream.itag == item.itag:
                return True
        return False

    def add(self, value: AudioOrVideoStream):
        self._elements.add(value)

    def discard(self, value: AudioOrVideoStream):
        try:
            self._elements.remove(value)
        except KeyError:
            pass

    def get_by_itag(self, itag: str) -> AudioOrVideoStream | None:
        for stream in self:
            if stream.itag == itag:
                return stream
        return None

    def filter(self, predicate: Callable[[AudioOrVideoStream], bool]) -> SetOfStreams:
        return self.__class__(list(filter(predicate, self._elements)))

    def query(
        self, format_spec: str, aliases: dict[str, str] | None = None
    ) -> list[AudioOrVideoStream]:
        if not format_spec:
            return []

        functions_map = {
            "best": lambda streams: sorted(streams, key=attrgetter("quality"))[-1],
        }

        if matched := FORMAT_SPEC_RE.search(format_spec):
            if function_name := matched.group("function"):
                expression = matched.group("expr")
            else:
                expression = matched.group("just_expr")
        else:
            raise QueryError(f"Format spec is invalid: {format_spec}")

        expression_filter = make_filter_from_expression(expression, aliases)
        queried: list[AudioOrVideoStream] = list(
            filter(expression_filter, self._elements)
        )
        if queried:
            if function_name:
                try:
                    queried = [functions_map[function_name](queried)]
                except KeyError:
                    raise QueryError(f"Unknown function: {function_name}")

        return queried



import copy
from dataclasses import asdict
from unittest.mock import patch
from urllib.parse import urljoin

import pytest
import responses
from ytpb.exceptions import MaxRetryError

from ytpb.playback import Playback
from ytpb.streams import AudioStream, Streams


class SideEffectCycler:
    def __init__(self, side_effects):
        self._side_effects = iter(side_effects)

    def __call__(self, *args, **kwargs):
        return next(self._side_effects)(*args, **kwargs)


@pytest.fixture()
def make_refresh_base_url_side_effect(streams_in_list, active_live_video_info):
    def wrapper(itag: str, new_base_url: str):
        def side_effect(obj: Playback, *args, **kwargs):
            updated_streams = copy.deepcopy(streams_in_list)
            for i, stream in enumerate(updated_streams):
                if stream.itag == itag:
                    stream_attributes = asdict(stream)
                    stream_attributes["base_url"] = new_base_url
                    updated_streams[i] = AudioStream(**stream_attributes)
                    break
            obj._streams = Streams(updated_streams)
            obj._info = active_live_video_info

        return side_effect

    return wrapper


def test_retry_on_403_for_segment_url(
    mocked_responses: responses.RequestsMock,
    mock_fetch_and_set_essential,
    make_refresh_base_url_side_effect,
    stream_url: str,
    audio_base_url: str,
) -> None:
    # Given:
    initial_segment_url = urljoin(audio_base_url, "sq/0")
    mocked_responses.get(initial_segment_url, status=403)

    refreshed_base_url = "https://test.googlevideo.com/videoplayback/test/"
    refreshed_segment_url = urljoin(refreshed_base_url, "sq/0")
    mocked_responses.get(refreshed_segment_url, status=200)

    # When:
    playback = Playback(stream_url)
    playback.fetch_and_set_essential()

    with patch.object(Playback, "fetch_and_set_essential", autospec=True) as mock:
        mock.side_effect = make_refresh_base_url_side_effect("140", refreshed_base_url)
        response = playback.session.get(initial_segment_url)

    # Then:
    assert response.status_code == 200
    assert response.url == refreshed_segment_url


def test_retry_on_404_for_segment_url(
    mocked_responses: responses.RequestsMock,
    mock_fetch_and_set_essential,
    stream_url: str,
    audio_base_url: str,
) -> None:
    # Given:
    initial_segment_url = urljoin(audio_base_url, "sq/0")
    mocked_responses.get(initial_segment_url, status=404)
    mocked_responses.get(initial_segment_url, status=200)

    # When:
    playback = Playback(stream_url)
    playback.fetch_and_set_essential()

    response = playback.session.get(initial_segment_url)

    # Then:
    assert response.status_code == 200
    assert response.url == initial_segment_url


def test_retry_on_unknown_for_segment_url(
    mocked_responses: responses.RequestsMock,
    mock_fetch_and_set_essential,
    stream_url: str,
    audio_base_url: str,
) -> None:
    # Given:
    initial_segment_url = urljoin(audio_base_url, "sq/0")
    mocked_responses.get(initial_segment_url, status=401)

    # When:
    playback = Playback(stream_url)
    playback.fetch_and_set_essential()
    response = playback.session.get(initial_segment_url)

    # Then:
    assert response.status_code == 401


def test_max_retries_exceeded_with_segment_url(
    mocked_responses: responses.RequestsMock,
    mock_fetch_and_set_essential,
    stream_url: str,
    audio_base_url: str,
) -> None:
    segment_url = urljoin(audio_base_url, "sq/0")
    mocked_responses.get(segment_url, status=403)

    playback = Playback(stream_url)
    playback.fetch_and_set_essential()

    with pytest.raises(MaxRetryError) as exc_info:
        playback.session.get(segment_url)

    assert exc_info.value.response.status_code == 403


def test_retry_head_request_with_segment_base_url(
    mocked_responses: responses.RequestsMock,
    mock_fetch_and_set_essential,
    make_refresh_base_url_side_effect,
    audio_base_url: str,
    stream_url: str,
) -> None:
    # Given:
    refreshed_base_url = "https://test.googlevideo.com/videoplayback/test/"
    mocked_responses.head(audio_base_url, status=403)
    mocked_responses.head(refreshed_base_url, status=200)

    # When:
    playback = Playback(stream_url)
    playback.fetch_and_set_essential()

    with patch.object(Playback, "fetch_and_set_essential", autospec=True) as mock:
        mock.side_effect = make_refresh_base_url_side_effect("140", refreshed_base_url)
        response = playback.session.head(audio_base_url)

    # Then:
    assert response.status_code == 200
    assert response.url == refreshed_base_url



from datetime import timezone
from pathlib import Path

import pytest

from ytpb.ffmpeg import ffprobe_show_entries


def assert_number_of_streams(filepath: Path, expected: int):
    assert ffprobe_show_entries(filepath, "format=nb_streams") == str(expected)


def assert_approx_duration(filepath: Path, expected: float, abs: float = 2.2e-2):
    actual = float(ffprobe_show_entries(filepath, "format=duration"))
    approx_expected = pytest.approx(expected, abs=abs)
    if actual != approx_expected:
        raise AssertionError(
            "Durations are not almost equal\n"
            f" - expected: {approx_expected}\n"
            f" + actual: {actual}"
        )


# See https://github.com/spulec/freezegun/issues/348.
def patched_freezgun_astimezone(self, tz=None):
    from freezegun.api import datetime_to_fakedatetime, real_datetime

    return datetime_to_fakedatetime(
        real_datetime.astimezone(self, timezone(self._tz_offset()))
    )



import io
from dataclasses import dataclass
from functools import total_ordering

from lxml import etree


NAMESPACES = {
    "mpd": "urn:mpeg:DASH:schema:MPD:2011",
    "yt": "http://youtube.com/yt/2012/10/10",
}


@total_ordering
@dataclass(frozen=True)
class VideoQuality:
    height: int
    frame_rate: float

    @classmethod
    def from_string(cls, value: str) -> "VideoQuality":
        try:
            height, frame_rate = value.split("p")
            return cls(int(height), float(frame_rate or 30))
        except ValueError:
            raise ValueError("Value not formatted as video quality")

    def __str__(self) -> str:
        if self.frame_rate == 30:
            return f"{self.height}p"
        else:
            return f"{self.height}p{self.frame_rate:.2g}"

    def __eq__(self, other) -> bool:
        return str(self) == str(other)

    def __gt__(self, other) -> bool:
        if isinstance(other, str):
            other = type(self).from_string(other)
        if self.height > other.height:
            return True
        elif self.height == other.height and self.frame_rate > other.frame_rate:
            return True
        else:
            return False


@dataclass(frozen=True, slots=True)
class RepresentationInfo:
    itag: str
    mime_type: str
    codecs: str
    base_url: str

    @property
    def format(self):
        "An alias for a MIME subtype."
        return self.mime_type.split("/")[1]

    def __repr__(self):
        return f"{type(self).__name__}(itag={self.itag})"


@dataclass(frozen=True, repr=False)
class AudioRepresentationInfo(RepresentationInfo):
    audio_sampling_rate: int


@dataclass(frozen=True, repr=False)
class VideoRepresentationInfo(RepresentationInfo):
    width: int
    height: int
    frame_rate: int

    @property
    def quality(self) -> VideoQuality:
        return VideoQuality(self.height, self.frame_rate)


def _eval_local_xpath(element: etree.Element, node: str) -> list[etree.Element]:
    return element.xpath(".//*[local-name() = $node]", node=node)


def strip_manifest(manifest: etree.Element) -> bytes:
    with io.BytesIO() as output_stream:
        etree.strip_elements(manifest, f"{{{NAMESPACES['mpd']}}}SegmentList")
        manifest.write(output_stream, pretty_print=True)
        return output_stream.getvalue()


def extract_representations_info(manifest_content: str) -> list[RepresentationInfo]:
    representations_info: list[RepresentationInfo] = []

    manifest = etree.fromstring(manifest_content.encode())

    adaptation_sets = manifest.xpath("//mpd:AdaptationSet", namespaces=NAMESPACES)
    for adaptation in adaptation_sets:
        mime_type = adaptation.get("mimeType")
        is_audio = "audio" in mime_type

        representations = _eval_local_xpath(adaptation, "Representation")
        for repr_ in representations:
            itag = repr_.get("id")
            base_repr_kwargs = dict(
                itag=itag,
                mime_type=mime_type,
                codecs=repr_.get("codecs"),
                base_url=_eval_local_xpath(repr_, "BaseURL")[0].text,
            )

            if is_audio:
                info = AudioRepresentationInfo(
                    **base_repr_kwargs,
                    audio_sampling_rate=int(repr_.get("audioSamplingRate")),
                )
            else:
                info = VideoRepresentationInfo(
                    **base_repr_kwargs,
                    width=int(repr_.get("width")),
                    height=int(repr_.get("height")),
                    frame_rate=int(repr_.get("frameRate")),
                )
            representations_info.append(info)

    return representations_info
